#! /usr/bin/env python
## Program to assist in understanding the series of interactions (C/C++)
## Dyncomp observes when performing comparability analysis.

## This takes in a trace file from a Dyncomp run  on a program and
## generates a graph of all the merges performed by Dyncomp. It then,
## using Python's built in pickle serialization, writes out the graph
## for use with merge_explorer

## TODO: It would be nice to track some data flow information on tags
## so a user can see where a tag is generated. Unfortunately Dyncomp 
## simply does not keep track of enough information for this.

## Usage merge_tracker trace-file
import logging;
import cPickle
import re, sys, pprint
import merge_utils
from pprint import pprint

if len(sys.argv) < 3:
  sys.exit("Usage: merge_tracker input-file output-db")

tag_source = {}
merge_graph = merge_utils.MergeGraph()
var_records = {}
var_records_set = {}

if sys.argv[1] is "-":
  print "Reading from standard input."
  f = sys.stdin
else:
  print "Opening file" + sys.argv[1]
  f = open(sys.argv[1])

output_file = open(sys.argv[2], "wb")

create_tag_regex     = "\[Dyncomp\] Creating fresh tag (.+?) at (.+) \((.+)\)"
value_merge_regex    = "\[Dyncomp\] Merging (.+?) with (.+?) to get (.+?) at (.+?)\)$"
obs_point_regex      = "\[Dyncomp\] OBSERVATION POINT: (.+?) - (.+?) \((.+?) (.+?) invocation (.+?)\)"

create_tag_re     = re.compile(create_tag_regex)
value_merge_re    = re.compile(value_merge_regex)
obs_point_re      = re.compile(obs_point_regex)

## Parse a line from the trace and update
## the interaction graph if necessary
def insertLine(line, regex, isVariable):
  match = regex.match(line)
  if match:
    tag1 = match.group(1);
    tag2 = match.group(2);
    sourceInfo = match.group(4)
    function = None

    if isVariable:
      function = match.group(5)
    merge_graph.insertEdge(merge_utils.MergeEdge(int(tag1), int(tag2), sourceInfo, function))    

# Main script begins here.

for line in f:
  # Update graph with tag creation. This just adds an unconnected
  # node and is not terribly useful without additional source
  # information
  create_match = create_tag_re.match(line)
  if create_match:
    (tag, source) = (int(create_match.group(1)), create_match.group(3))
    tag_source[tag] = source
    continue

  # Update graph with an observation point. This represents Dyncomp
  # updating it's knowledge of values held by a particular variable.
  obs_point_match = obs_point_re.match(line)
  if obs_point_match:
    isGlobal = False
    (var, tag, entryInfo, func, invocation) =  (obs_point_match.group(1),
                                                obs_point_match.group(2),
                                                obs_point_match.group(3),
                                                obs_point_match.group(4),
                                                obs_point_match.group(5))
    is_entry = (entryInfo == "ENTRY")

    if var[0] == "/":
      isGlobal = True

    if not isGlobal:
      unique_name = var.strip() + "." + func.strip()
    else:
      unique_name = var.strip()      

    tag = int(tag)
    invocation = int(invocation)
    if unique_name not in var_records:
      var_records[unique_name] = merge_utils.VarRecord()
    var_records[unique_name].tags.append(merge_utils.ObservationPoint(tag, func, is_entry, invocation))
  
  # Handle a possible interaction line.
  insertLine(line, value_merge_re   , False)

print "Creating merge database."  
cPickle.dump(var_records, output_file, -1)
cPickle.dump(merge_graph, output_file, -1)
cPickle.dump(tag_source, output_file, -1)
output_file.close()
